import numpy as np

from depynd.information import mutual_information


def _mrmr(X, y, lamb, k, **kwargs):
    """Select effective features in ``X`` on predicting ``y`` using minimum redundancy maximum relevance feature
    selection [peng2005feature]_.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Observations of feature variables.
    y : array-like, shape (n_samples)
        Observations of the target variable.
    lamb: float or None
        Threshold for independence tests. Ignored if `k` is specified.
    k : int or None
        Number of selected features.
    kwargs : dict
        Optional parameters for MI estimation.

    Returns
    -------
    indices : list
        Indices of the selected features.

    References
    ----------
    .. [peng2005feature] Peng, Hanchuan, Fuhui Long, and Chris Ding. "Feature selection based on mutual information
        criteria of max-dependency, max-relevance, and min-redundancy." IEEE Transactions on pattern analysis and
        machine intelligence 27.8 (2005): 1226-1238.
    """
    n, d = X.shape
    if k is not None:
        return _grow([], X, y, -np.inf, k, **kwargs)
    else:
        return _grow([], X, y, lamb, d, **kwargs)


def _grow(selected, X, y, lamb, k, **kwargs):
    n, d = X.shape
    while True:
        max_obj = -np.inf
        not_selected = set(range(d)) - set(selected)
        for i in not_selected:
            x = X[:, i]
            rel = mutual_information(x, y, **kwargs)
            red = [mutual_information(x, X[:, j], **kwargs) for j in selected]
            obj = rel - (np.mean(red) if red else 0)
            if max_obj < obj:
                max_obj = obj
                max_idx = i
        if max_obj <= lamb or len(selected) == k:
            return selected
        selected.append(max_idx)



import numpy as np
from sklearn.preprocessing import scale


def _instability(X, estimator, lamb, ratio, rep_num, **kwargs):
    n, p = X.shape
    b = int(ratio * n)
    indices = [np.random.choice(np.arange(n), size=b) for _ in range(rep_num)]
    samples = [scale(sample) for sample in X[indices, :]]
    adjs = [estimator(sample, lamb, **kwargs) for sample in samples]
    theta = np.sum(adjs, axis=0) / rep_num
    xi = 2 * theta * (1 - theta)
    d = np.sum(xi) / p / (p - 1)
    return d


def _stars(X, estimator, lamb, beta, ratio, rep_num, verbose=False, **kwargs):
    """Obtain the best regularization parameter using Stability Approach to Regularization Selection
    [liu2010stability]_.

    Parameters
    ----------
    X : array, shape (n_samples, n_features)
        Observations of variables.
    estimator
    lamb
    beta
    ratio
    rep_num
    verbose

    Returns
    -------
    lambda_opt : float
        Optimal regularization parameter.

    References
    -------
    .. [liu2010stability] Liu, Han, Kathryn Roeder, and Larry Wasserman. "Stability approach to regularization selection
        (stars) for high dimensional graphical models." Advances in neural information processing systems. 2010.
    """
    for i, l in enumerate(lamb):
        instability = _instability(X, estimator, l, ratio=ratio, rep_num=rep_num, **kwargs)
        if instability > beta:
            return lamb[i - 1]
        if verbose:
            print('[stars] lambda: %f, instability: %f' % (lamb, instability))
    return 0



import numpy as np
from sklearn.utils import check_array

from depynd.markov_networks import _skeptic, _stars, _glasso, _gsmn, _iamb, _gsmple


def select(X, method='skeptic', criterion='stars', lamb=None, verbose=False, return_lambda=False, **kwargs):
    """Learn the structure of Markov random field.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Observations of a set of random variables.
    method : {'glasso', 'skeptic', 'gsmn', 'iamb'}
        Method for structure learning.
    criterion : {'stars', 'none'}
        Criteria for selecting regularization parameter.
    lamb : float or array-like or None
        Candidates of regularization parameter.
    verbose : bool
        If True, the objective function is plotted for each regularization parameter.
    return_lambda : bool, default False
        If True, the selected regularization parameter will be returned.
    kwargs : dict
        Optional parameters for MI estimation.

    Returns
    -------
    adj : array, shape (n_features, n_features)
        Estimated adjacency matrix of an MRF.
    """
    X = check_array(X, ensure_min_samples=2, ensure_min_features=2)

    if lamb is None:
        lamb = [1e-5, 1e-4, 1e-3, 5e-3, 0.01, 0.03, 0.05, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1]

    if method == 'glasso':
        estimator = _glasso
        _check_lamb(lamb, check_non_negative=True, method=method)
    elif method == 'skeptic':
        estimator = _skeptic
        _check_lamb(lamb, check_non_negative=True, method=method)
    elif method == 'gsmn':
        estimator = _gsmn
        _check_lamb(lamb, check_non_negative=False, method=method)
    elif method == 'iamb':
        estimator = _iamb
        _check_lamb(lamb, check_non_negative=False, method=method)
    elif method == 'gsmple':
        estimator = _gsmple
        _check_lamb(lamb, check_non_negative=False, method=method)
    else:
        raise ValueError('`%s` is not implemented.' % method)

    if criterion == 'none':
        if np.iterable(lamb):
            lamb_opt = next(iter(lamb))
        elif np.isscalar(lamb):
            lamb_opt = lamb
        else:
            raise ValueError('`lamb` should be None, a scalar, or an iterable of scalars.')
    elif criterion == 'stars':
        if np.iterable(lamb) and len(lamb) > 1:
            lamb = sorted(lamb, reverse=True)  # sort by descending order
            if 'beta' not in kwargs:
                kwargs['beta'] = 0.1
            if 'ratio' not in kwargs:
                n = len(X)
                kwargs['ratio'] = 10 * (n ** -0.5) if n > 144 else 0.8
            if 'rep_num' not in kwargs:
                kwargs['rep_num'] = 20
            print(lamb)
            lamb_opt = _stars(X, estimator, lamb=lamb, verbose=verbose, **kwargs)
        elif np.iterable(lamb) and len(lamb) == 1:
            lamb_opt = next(iter(lamb))
        elif np.isscalar(lamb):
            lamb_opt = lamb
        else:
            raise ValueError('`lamb` should be None, a scalar, or an iterable of scalars.')
    else:
        raise ValueError('Criteria %s is not implemented.' % criterion)

    if return_lambda:
        return estimator(X, lamb_opt), lamb_opt
    else:
        return estimator(X, lamb_opt)


def _check_lamb(lamb, check_non_negative, method):
    if np.iterable(lamb):
        if not all(np.isscalar(l) for l in lamb):
            raise ValueError('`lamb` must be None, a scalar, or an iterable of scalars.')
        if check_non_negative and any(l < 0 for l in lamb):
            raise ValueError('Every element in `lamb` must be positive when using %s.' % method)
    else:
        if not np.isscalar(lamb):
            raise ValueError('`lamb` must be None, a scalar, or an iterable of scalars.')
        if check_non_negative and lamb < 0:
            raise ValueError('`lamb` must be positive when using %s.' % method)



import numpy as np


def _mi_plugin(X, Y):
    XY = np.hstack([X, Y])
    return _entropy_plugin(X) + _entropy_plugin(Y) - _entropy_plugin(XY)


def _entropy_plugin(X):
    n, d = X.shape
    _, cnt = np.unique(X, axis=0, return_counts=True)
    p = cnt / n
    return -np.sum(p * np.log(p))



from ._jose import _jose
from ._skeptic import _skeptic
from ._glasso import _glasso
from ._gsmn import _gsmn
from ._iamb import _iamb
from ._gsmple import _gsmple
from ._stars import _stars
from ._select import select

__all__ = ['_jose', '_skeptic', '_glasso', '_gsmple', '_gsmn', '_iamb', '_stars', 'select']



from itertools import combinations

import numpy as np
from sklearn.linear_model import Lasso


def non_diag(A):
    tmp = np.copy(A)
    tmp[np.eye(len(A), dtype=bool)] = 0
    return tmp


def _jose(X, lamb=0.01, tol=1e-6, max_iter=100, **kwargs):
    if set(np.ravel(X)) != {0, 1}:
        raise ValueError('Each element of X must be in {0, 1}.')

    n, d = X.shape
    comb = list(combinations(np.arange(d), 2))
    theta = np.eye(d)
    for _ in range(max_iter):
        theta_prev = np.copy(theta)
        z = np.diag(theta)[None, :] + X @ non_diag(theta)
        p = 1 / (1 + np.exp(-z))
        w = p * (1 - p)
        y = np.diag(theta)[None, :] + X @ non_diag(theta) - (p - X) / w
        y_star = np.sqrt(w) * y
        y_mean = np.mean(y_star, axis=0)
        y_star_star = y_star - y_mean[None, :]
        x_star = np.sqrt(w) * X
        x_mean = np.mean(x_star, axis=0)
        x_star_star = x_star - x_mean[None, :]
        x = np.zeros([n * d, d * (d - 1) // 2])
        for k, (i, j) in enumerate(comb):
            x[np.arange(n * i, n * (i + 1)), k] = x_star_star[:, j]
            x[np.arange(n * j, n * (j + 1)), k] = x_star_star[:, i]
        y = np.concatenate(y_star_star.T)
        coef = Lasso(lamb).fit(x, y).coef_
        for k, (i, j) in enumerate(comb):
            theta[i, j] = theta[j, i] = coef[k]
        w_mean = np.mean(np.sqrt(w), axis=0)
        theta[np.eye(d, dtype=bool)] = (y_mean - x_mean @ non_diag(theta)) / w_mean
        diff = np.linalg.norm(theta - theta_prev)
        if diff < tol:
            break

    adj = np.abs(theta) > 1e-6
    adj[np.eye(d, dtype=bool)] = 0
    return adj



import numpy as np

from depynd.information import conditional_mutual_information


def _iamb(X, lamb=0.0, **kwargs):
    """Learn the structure of Markov random field by finding Markov blanket for each variable with Incremental
    Association Markov Blanket [tsamardinos2003algorithms]_.

    Parameters
    ----------
    X : array, shape (n_samples, n_features)
        Observations of variables.
    lamb: float
        Threshold for independence test.
    kwargs : dict, default None
        Optional parameters for MI estimation.

    Returns
    -------
    adj : array, shape (n_features, n_features)
        Estimated adjacency matrix of an MRF.

    References
    ----------
    .. [tsamardinos2003algorithms] Tsamardinos, Ioannis, et al. "Algorithms for Large Scale Markov Blanket Discovery."
        FLAIRS conference. Vol. 2. 2003.
    """
    n, d = X.shape
    adj = np.zeros([d, d], dtype=bool)
    for i in range(d):
        adj_tmp = np.zeros([d, d], dtype=bool)
        adj_tmp = _grow(adj_tmp, i, X, lamb, **kwargs)
        adj_tmp = _shrink(adj_tmp, i, X, lamb, **kwargs)
        adj |= adj_tmp
    return adj


def _grow(adj, i, X, lamb, **kwargs):
    n, d = X.shape
    x = X[:, i]
    updated = True
    while updated:
        updated = False
        vmax = -np.inf
        z = X[:, adj[i]]
        non_adj = ~adj[i] & (np.arange(d) != i)
        for j in non_adj.nonzero()[0]:
            y = X[:, j]
            cmi = conditional_mutual_information(x, y, z, **kwargs)
            if vmax < cmi:
                vmax = cmi
                imax, jmax = i, j
        if vmax >= lamb:
            adj[imax, jmax] = adj[jmax, imax] = 1
            updated = True
    return adj


def _shrink(adj, i, X, lamb, **kwargs):
    n, d = X.shape
    x = X[:, i]
    for j in adj[i].nonzero()[0]:
        other_adj = adj[i] & (np.arange(d) != j)
        y = X[:, j]
        z = X[:, other_adj]
        cmi = conditional_mutual_information(x, y, z, **kwargs)
        if cmi <= lamb:
            adj[i, j] = adj[j, i] = 0
        if np.count_nonzero(adj[i]) == 0:
            break
    return adj



import numpy as np
from sklearn.utils.validation import check_array

from depynd.information import _mi_dr, _mi_knn, _mi_plugin


def mutual_information(X, Y, mi_estimator='auto', is_discrete='auto', force_non_negative=False, **kwargs):
    """Estimate mutual information between ``X`` and ``Y``.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features_x) or (n_samples)
        Observations of a variable.
    Y : array-like, shape (n_samples, n_features_y) or (n_samples)
        Observations of the other variable.
    mi_estimator : {'knn', 'dr', 'plugin', 'auto'}, default 'auto'
        MI estimator. If 'auto', MI estimator will be selected depending on whether all features are purely discrete,
        purely continuous, or discrete-continuous mixed. If purely discrete, 'plugin' estimator will be used. Otherwise,
        'knn' estimator will be selected.
    is_discrete : {'auto', bool}, default 'auto'
        If ``bool``, then it determines whether to consider all features purely discrete or purely continuous. If
        'auto', a column which contains duplicate elements will be considered discrete.
    force_non_negative : bool, default False
        If ``True``, the result will be taken max with zero.
    kwargs : dict
        Optional parameters for MI estimation.

    Returns
    -------
    mi : float
        Estimated mutual information between ``X`` and ``Y``.
    """
    if np.size(X) == 0 or np.size(Y) == 0:
        return 0
    X = np.atleast_2d(X.T).T
    Y = np.atleast_2d(Y.T).T
    X = check_array(X, ensure_min_samples=2)
    Y = check_array(Y, ensure_min_samples=2)
    assert len(X) == len(Y), 'X and Y must have the same length.'

    n = len(X)
    if is_discrete == 'auto':
        is_discrete = all(n != len(set(col)) for col in X.T) and all(n != len(set(col)) for col in Y.T)
        is_continuous = all(n == len(set(col)) for col in X.T) and all(n == len(set(col)) for col in Y.T)
    elif isinstance(is_discrete, bool):
        is_continuous = not is_discrete and \
                        all(n == len(set(col)) for col in X.T) and all(n == len(set(col)) for col in Y.T)
    else:
        raise TypeError("`is_discrete` must be 'auto' or bool.")

    if mi_estimator == 'auto':
        if is_discrete:
            mi_estimator = 'plugin'
        # elif is_continuous:  # unstable
        #     mi_estimator = 'dr'
        else:
            mi_estimator = 'knn'

    if mi_estimator == 'dr':
        sigma = kwargs.get('sigma', 1)
        n_bases = kwargs.get('n_bases', 200)
        maxiter = kwargs.get('maxiter', 1000)
        assert sigma > 0, '`sigma` must be positive.'
        assert isinstance(n_bases, (int, np.integer)) and n_bases > 0, '`n_bases` must be a positive integer.'
        assert isinstance(maxiter, (int, np.integer)) and maxiter > 0, '`maxiter` must be a positive integer.'
        assert is_continuous, 'When using density ratio estimator, all features must be continuous.'
        mi = _mi_dr(X, Y, sigma=sigma, n_bases=n_bases, maxiter=maxiter)
    elif mi_estimator == 'knn':
        n_neighbors = kwargs.get('n_neighbors', 3)
        assert isinstance(n_neighbors, (int, np.integer)), '`n_neighbors` must be an integer.'
        assert n_neighbors > 0, '`n_neighbors` must be positive.'
        assert n_neighbors < len(X), '`n_neighbors` must be smaller than `n_sample`.'
        mi = _mi_knn(X, Y, n_neighbors)
    elif mi_estimator == 'plugin':
        assert is_discrete, 'When using plug-in estimator, all features must be discrete.'
        mi = _mi_plugin(X, Y)
    else:
        raise ValueError('`%s` is not implemented.' % mi_estimator)

    return max(mi, 0) if force_non_negative else mi


def conditional_mutual_information(X, Y, Z, mi_estimator='auto', is_discrete='auto', force_non_negative=False,
                                   **kwargs):
    """Estimate conditional mutual information between ``X`` and ``Y`` given ``Z``.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features_x) or (n_samples)
        Observations of a conditioned variable.
    Y : array-like, shape (n_samples, n_features_y) or (n_samples)
        Observations of the other conditioned variable.
    Z : array-like, shape (n_samples, n_features_z) or (n_samples)
        Observations of the conditioning variable.
    mi_estimator : {'knn', 'dr', 'plugin', 'auto'}, default 'auto'
        MI estimator. If 'auto', MI estimator will be selected depending on whether all features are purely discrete,
        purely continuous, or discrete-continuous mixed. If purely discrete, 'plugin' estimator will be used. Otherwise,
        'knn' estimator will be selected.
    is_discrete : {'auto', bool}, default 'auto'
        If ``bool``, then it determines whether to consider all features purely discrete or purely continuous. If
        'auto', a column which contains duplicate elements will be considered discrete.
    force_non_negative : bool, default False
        If ``True``, the result will be taken max with zero.
    kwargs : dict, default None
        Optional parameters for MI estimation.

    Returns
    -------
    cmi : float
        Estimated conditional mutual information between ``X`` and ``Y``, given ``Z``.
    """
    if np.size(Z) == 0:
        return mutual_information(X, Y, mi_estimator, is_discrete, force_non_negative, **kwargs)
    assert len(X) == len(Y) == len(Z), 'X, Y and Z must have the same length.'
    X = np.atleast_2d(X.T).T
    Z = np.atleast_2d(Z.T).T
    XZ = np.hstack([X, Z])
    mi_xz_y = mutual_information(XZ, Y, mi_estimator, is_discrete, **kwargs)
    mi_y_z = mutual_information(Y, Z, mi_estimator, is_discrete, **kwargs)
    cmi = mi_xz_y - mi_y_z
    return max(cmi, 0) if force_non_negative else cmi



import numpy as np

from depynd.information import conditional_mutual_information


def _gsmple(X, lamb=0.0, **kwargs):
    n, d = X.shape
    adj = np.zeros([d, d], dtype=bool)
    adj = _grow(adj, X, lamb, **kwargs)
    adj = _shrink(adj, X, lamb, **kwargs)
    return adj


def _grow(adj, X, lamb, **kwargs):
    # TODO: CMI caching
    n, d = X.shape
    while True:
        vmax = -np.inf
        for i in range(d):
            x = X[:, [i]]
            z = X[:, adj[i]]
            non_adj = ~adj[i] & (np.arange(d) != i)
            for j in non_adj.nonzero()[0]:
                if i <= j:
                    continue
                y = X[:, [j]]
                w = X[:, adj[j]]
                cmi = conditional_mutual_information(x, y, z, **kwargs)
                cmi += conditional_mutual_information(x, y, w, **kwargs)
                if vmax < cmi:
                    vmax = cmi
                    imax, jmax = i, j

        if vmax <= lamb:
            return adj

        adj[imax, jmax] = adj[jmax, imax] = 1

        if np.count_nonzero(adj) == d ** 2 - d:
            return adj


def _shrink(adj, X, lamb, **kwargs):
    # TODO: CMI caching
    n, d = X.shape
    while True:
        vmin = np.inf
        for i in range(d):
            x = X[:, [i]]
            for j in adj[i].nonzero()[0]:
                if i <= j:
                    continue
                other_adj_i = adj[i] & (np.arange(d) != j)
                other_adj_j = adj[j] & (np.arange(d) != i)
                y = X[:, [j]]
                z = X[:, other_adj_i]
                w = X[:, other_adj_j]
                cmi = conditional_mutual_information(x, y, z, **kwargs)
                cmi += conditional_mutual_information(x, y, w, **kwargs)
                if vmin > cmi:
                    vmin = cmi
                    imin, jmin = i, j

        if vmin > lamb:
            return adj

        adj[imin, jmin] = adj[jmin, imin] = 0

        if np.count_nonzero(adj) == 0:
            return adj



from ._dr import _mi_dr
from ._knn import _mi_knn
from ._plugin import _entropy_plugin, _mi_plugin
from ._information import mutual_information, conditional_mutual_information

__all__ = ['_mi_dr', '_mi_knn', '_entropy_plugin', '_mi_plugin', 'mutual_information',
           'conditional_mutual_information']



from setuptools import setup, find_packages

long_description = """depynd is a Python library for evaluating dependencies among random variables from data.
It supports learning statistical dependencies for one-to-one, one-to-many, and many-to-many relationships,
where each one corresponds to

- mutual information (MI) estimation,
- feature selection, and
- graphical model structure learning,

respectively. Specifically, depynd supports MI estimation for discrete-continuous mixtures, MI-based feature
selection, and structure learning of undirected graphical models (a.k.a. Markov random fields)."""

setup(
    name='depynd',
    version='0.6.2',
    description='Evaluating dependencies among random variables.',
    long_description=long_description,
    author='Yuya Takashina',
    author_email='takashina2051@gmail.com',
    packages=find_packages(),
    install_requires=['numpy>=1.13.0', 'scipy', 'scikit-learn'],
    test_requires=['pytest', 'flake8'],
    url='https://github.com/y-takashina/depynd',
)



import numpy as np

from depynd.information import conditional_mutual_information


def _mifs(X, y, lamb, k, **kwargs):
    """Select effective features in ``X`` on predicting ``y`` using mutual-information-based feature selection
    [brown2012conditional]_.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Observations of feature variables.
    y : array-like, shape (n_samples)
        Observations of the target variable.
    lamb: float or None
        Threshold for independence tests. Ignored if `k` is specified.
    k : int or None
        Number of selected features.
    kwargs : dict
        Optional parameters for MI estimation.

    Returns
    -------
    indices : list
        Indices of the selected features.

    References
    ----------
    .. [brown2012conditional] Brown, Gavin, et al. "Conditional likelihood maximisation: a unifying framework for
        information theoretic feature selection." Journal of machine learning research 13.Jan (2012): 27-66.
    """
    n, d = X.shape
    selected = []
    if k is not None:
        selected = _grow(selected, X, y, -np.inf, k, **kwargs)
    else:
        selected = _grow(selected, X, y, lamb, d, **kwargs)
        selected = _shrink(selected, X, y, lamb, 0, **kwargs)
    return selected


def _grow(selected, X, y, lamb, k, **kwargs):
    n, d = X.shape
    while True:
        max_cmi = -np.inf
        not_selected = set(range(d)) - set(selected)
        z = X[:, selected]
        for i in not_selected:
            x = X[:, i]
            cmi = conditional_mutual_information(x, y, z, **kwargs)
            if max_cmi < cmi:
                max_cmi = cmi
                max_idx = i
        if max_cmi <= lamb or len(selected) == k:
            return selected
        selected.append(max_idx)


def _shrink(selected, X, y, lamb, k, **kwargs):
    while True:
        min_cmi = np.inf
        for i in selected:
            x = X[:, i]
            z = X[:, list(set(selected) - {i})]
            cmi = conditional_mutual_information(x, y, z, **kwargs)
            if min_cmi > cmi:
                min_cmi = cmi
                min_idx = i
        if min_cmi >= lamb or len(selected) == k:
            return selected
        selected.remove(min_idx)



import numpy as np
from sklearn.utils.validation import check_X_y

from depynd.feature_selection import _mrmr, _mifs


def select(X, y, lamb=0.0, k=None, method='mifs', **kwargs):
    """Select effective features in ``X`` on predicting ``y``.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Observations of feature variables.
    y : array-like, shape (n_samples)
        Observations of the target variable.
    lamb: float or None
        Threshold for independence tests. Ignored if `k` is specified.
    k : int or None
        Number of selected features.
    method: {'mifs', 'mrmr'}, default 'mifs'
        Feature selection method.

    Returns
    -------
    indices : list
        Indices of the selected features.
    """
    y = np.ravel(y)
    X, y = check_X_y(X, y, ensure_min_samples=2, ensure_min_features=2)
    if lamb is None and k is None:
        raise ValueError('At least either `lamb` or `k` should be specified.')
    if k is not None:
        assert isinstance(k, (int, np.integer)) and k > 0, '`k` must be a positive integer.'
        assert k <= X.shape[1], '`k` cannot be larger than number of features.'
    else:
        assert np.isscalar(lamb), '`lamb` must be a real value.'
    if method == 'mifs':
        return _mifs(X, y, lamb=lamb, k=k, **kwargs)
    elif method == 'mrmr':
        return _mrmr(X, y, lamb=lamb, k=k, **kwargs)
    else:
        raise ValueError('`%s` is not implemented.' % method)



import numpy as np
from sklearn.preprocessing import scale
from sklearn.covariance import graph_lasso


def _glasso(X, lamb, return_precision=False, **kwargs):
    """Learn the structure of Markov random field with the graphical lasso.

    This function internally calls the implementation in scikit-learn.

    Parameters
    ----------
    X : array, shape (n_samples, n_features)
        Observations of variables.
    lamb : float
        Regularization parameter.
    return_precision : bool, default False
        If True, the estimated precision matrix will be returned instead of adjacency matrix.

    Returns
    ----------
    adj : array, shape (n_features, n_features)
        Estimated adjacency matrix (or precision matrix if ``return_precision`` is True) of an MRF.
    """
    cov = np.cov(scale(X), rowvar=False)
    pre = graph_lasso(cov, alpha=lamb)[1]
    if return_precision:
        return pre
    else:
        adj = ~np.isclose(pre, 0)
        adj[np.eye(len(adj), dtype=bool)] = 0
        return adj
